# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/cell2func.ipynb.

# %% auto 0
__all__ = ['get_non_callable_ipython', 'get_non_callable', 'FunctionProcessor', 'update_cell_code', 'add_function_to_list',
           'CellProcessor', 'CellProcessorMagic', 'load_ipython_extension', 'keep_variables',
           'keep_variables_in_memory', 'acceptable_variable', 'store_variables']

# %% ../../nbs/cell2func.ipynb 3
import pdb
import joblib
import os
import re
import argparse
import shlex
from dataclasses import dataclass
from functools import reduce
from pathlib import Path
import sys
import ast
import logging

from IPython import get_ipython
from IPython.core.magic import (Magics, magics_class, line_magic,
                                cell_magic, line_cell_magic)
from IPython.core.magic_arguments import (argument, magic_arguments, parse_argstring)
import ipynbname
from sklearn.utils import Bunch
from fastcore.all import argnames
import nbdev

# %% ../../nbs/cell2func.ipynb 7
import pdb
def get_non_callable_ipython (variables_to_inspect, locals_, self=None):
    """
    Store `variables` in dictionary entry `self.variables_field[function]`
    """
    frame_number = 0
    ##pdb.no_set_trace()
    
    while not isinstance (self, FunctionProcessor):
        try:
            fr = sys._getframe(frame_number)
        except:
            break
        args = argnames(fr, True)
        if len(args)>0:
            self = fr.f_locals[args[0]]
        frame_number += 1
        
    if isinstance (self, FunctionProcessor):
        non_callable_variables = f'{variables_to_inspect}_'
        self[non_callable_variables] = []
        for name in self[variables_to_inspect]:
            try:
                x = locals_[name]
                if not callable (x):
                    self[non_callable_variables].append(name)
            except: 
                self[non_callable_variables].append(name)
        self[variables_to_inspect] = self[non_callable_variables].copy()

# %% ../../nbs/cell2func.ipynb 8
def get_non_callable (variables):    
    non_callable=[]
    for name in variables:
        try:
            x = eval(name)
            if not callable (x):
                non_callable.append(name)
        except: 
            non_callable.append(name)
    return non_callable

# %% ../../nbs/cell2func.ipynb 10
class FunctionProcessor (Bunch):
    """
    Function processor.
    """
    def to_file (self, file_path, mode='w'):
        with open (file_path, mode=mode) as file:
            file.write (self.code)
    
    def write (self, file):
        file.write (self.code)
        file.write ('\n')
        
    def print (self):
        print (self.code)
        
    def __getattr__ (self, k):
        ##pdb.no_set_trace()
        try:
            v = super().__getattr__ (k)
            return v
        except:
            if 'current_values' in self.keys() and k in self.current_values:
                return self.current_values[k]
            else:
                return None
    
    def update_code (
        self, 
        arguments=None, 
        return_values=None,
        display=False
    ) -> None:
        #pdb.no_set_trace()
        if self.permanent:
            get_ipython().run_cell(self.code)
            return
        if self.data:
            arguments = []
            arguments = 'test=False'
        else:
            if arguments is not None:
                arguments += [k.split('=')[0] for k in self.include_input if k not in arguments]
                arguments = [k for k in arguments if k not in self.exclude_input]
                self.arguments = arguments
            arguments = ', '.join (self.arguments)
        unpack_input_code=''
        pack_output_code=''
        bunch_variable = None
        if self.unpack_bunch is not None:
            bunch_name = self.unpack_bunch
            bunch_variable = self.previous_values[bunch_name] if bunch_name in self.previous_values else self.current_values[bunch_name]
            if not isinstance (bunch_variable, dict):
                raise ValueError (f'{bunch_name} not found, or not instance of dict')
            unpack_input_code = ''.join ([f'{" " * self.tab_size}{k} = {bunch_name}["{k}"]\n' for k in bunch_variable])
            if self.pack_all:
                pack_output_code = ''.join ([f'{" " * self.tab_size}{bunch_name}["{k}"] = {k}\n' for k in bunch_variable])
            else:
                pack_output_code = ''.join ([f'{" " * self.tab_size}{bunch_name}["{k}"] = {k}\n' for k in bunch_variable if k in self.created_variables ])

        if self.return_all is None: self.return_all = False
        if self.only_posterior is None: self.only_posterior = False
        if self.return_all and self.only_posterior:
            raise ValueError ('only one of return-all or only-posterior can be indicated')
        if self.return_all or (self.unpack_bunch and not self.only_posterior):
            return_values = self.created_variables
        if return_values is not None:
            return_values += [k for k in self.include_output if k not in return_values]
            return_values = [k for k in return_values if k not in self.exclude_output]
            if bunch_variable is not None:
                new_return_values = [k for k in return_values if k not in bunch_variable]
                new_return_values += [k.split('=')[0] for k in self.include_input if k not in bunch_variable]
                pack_output_code += ''.join ([f'{" " * self.tab_size}{bunch_name}["{k}"] = {k}\n' for k in new_return_values ])
                return_values = [bunch_name]
            self.return_values = return_values
        elif self.unpack_bunch is not None:
            self.return_values = [bunch_name]
            
        if self.returns_dict:
            return_values = 'dict (\n' + ''.join([f'{x}={x},\n' for x in self.return_values]) + ')'
        if self.returns_bunch:
            return_values = 'Bunch (\n' + ''.join([f'{x}={x},\n' for x in self.return_values]) + ')'
        else:
            return_values = ','.join (self.return_values)
        function_code = ''
        tab_size =  self.tab_size if not self.defined else 0
        for line in self.original_code.splitlines():
            if 'if True:' not in line:
                function_code += f'{" " * tab_size}{line}\n'
        if return_values != '':
            return_line = f'return {return_values}'
            return_line = f'{" " * self.tab_size}{return_line}\n'
        else:
            return_line = ''
        function_calls = '' if 'function_calls' not in self else self.function_calls
        signature = self.signature if self.signature is not None else f'def {self.name}({arguments}):'
        # store parts
        self.code_parts=Bunch(
            signature=signature, 
            function_calls=function_calls, 
            unpack_input_code=unpack_input_code, 
            function_code=function_code, 
            pack_output_code=pack_output_code,
            return_line=return_line
        )
        # assemble
        function_code = f'{signature}\n' + function_calls + unpack_input_code + function_code + pack_output_code + return_line
        self.code = function_code
        get_ipython().run_cell(function_code)
        param_assignment_code = (
f'''
import inspect
x=inspect.signature ({self.name})
params = dict(x.parameters)
for k, v in params.items():
    if (v.default) is not inspect._empty:
        exec (f'k = v.default')

'''
        )
        #get_ipython().run_cell(param_assignment_code)
        if display:
            print (function_code)
    
    def get_ast(self, original=True, code=None):
        if code is None:
            code = self.original_code if original else self.code
        print(ast.dump(ast.parse(code), indent=2))
        
    def parse_variables (self, code=None):
        if code is None: code=self.original_code
        # variable parsing
        root = ast.parse (code)
        
        # ---------------------------------------
        # newly created names: candidates for return list and not for argument list
        self.non_unique_created_variables = [node.id for node in ast.walk(root) if isinstance(node, ast.Name) and isinstance(node.ctx, ast.Store)]
        # avoid duplicates:
        self.created_variables = []
        for node_id in self.non_unique_created_variables:
            if node_id not in self.created_variables: self.created_variables.append(node_id)
        
        # ---------------------------------------
        # names defined before: candidates for arguments list, if they are not callable
        self.non_unique_loaded_names = [node.id for node in ast.walk(root) if isinstance(node, ast.Name) and isinstance(node.ctx, ast.Load)]
        # avoid duplicates:
        self.loaded_names = []
        for node_id in self.non_unique_loaded_names:
            if node_id not in self.loaded_names: self.loaded_names.append(node_id)
        
        self.previous_variables = [x for x in self.loaded_names if x not in self.created_variables]
        if self.unpack_bunch is not None and self.unpack_bunch not in self.previous_variables:
            self.previous_variables.append (self.unpack_bunch)
        
        # names that appear as arguments in functions -> some defined created the current function, some in the current one
        v=[node for node in ast.walk(root) if isinstance(node, ast.Call)]
        self.argument_variables = [y.id  for x in v for y in x.args if isinstance(y, ast.Name)]
        # argument variables might still be modified in the function, so they need to be marked as I/O, i.e., candidates for return list and for argument list
        
        # loaded names that are not arguments and not created in the current function are most probably read-only, i.e., not candidates for return list
        self.read_only_variables = [x for x in self.previous_variables if x not in self.argument_variables]
        self.posterior_variables = []
        self.all_variables = self.created_variables.copy()
        self.all_variables += [k for k in self.previous_variables if k not in self.all_variables]
        self.all_variables += [k for k in self.argument_variables if k not in self.all_variables]
        
    def parse_arguments_and_results (self, root):
        function_visitor = FunctionVisitor ()
        function_visitor.visit (root)
        return_visitor = FunctionVisitor ()
        return_visitor.visit (root)
        self.arguments = function_visitor.arguments
        self.return_values = function_visitor.return_values
        loaded_names_not_in_arguments = set(self.loaded_names).difference (self.arguments)
        if len(loaded_names_not_in_arguments) > 0:
            self.logger.debug (f'The following loaded names were not found in the arguments list: {loaded_names_not_in_arguments}')
            
    def _store_values (self, field, code="", store_values=True):
        """Stores local variables in field `field`of self"""
        code_to_run1 = code + f'from nbmodular.core.cell2func import keep_variables_in_memory\nkeep_variables_in_memory ("{field}", locals ())'
        code_to_run2 = code + f'from nbmodular.core.cell2func import keep_variables\nkeep_variables ("{field}", locals ())'
        get_ipython().run_cell (code_to_run1)
        if 'created_current_values' not in self[field]:
            self.logger.debug ('storing local variables in disk')
            get_ipython().run_cell (code_to_run2)
            self[field] = joblib.load ('variable_values.pk')
            os.remove ('variable_values.pk')
        else:
            del self[field]['created_current_values']
        
        if not store_values:
            self[field] = {k: '__REMOVED__' for k in self[field]}
        
    def _create_function_info_object (self):
        info_object_code=(
f'''
import joblib
from nbmodular.core.cell2func import FunctionProcessor
from nbmodular.core.cell2func import acceptable_variable

variable_values = locals()
variable_values = {{k: variable_values[k] for k in variable_values if acceptable_variable(variable_values, k)}}
{self.name}_info = joblib.load ('function_processor.pk')
{self.name}_info = FunctionProcessor (**{self.name}_info)
{self.name}_info.current_values = variable_values
''')
        get_ipython().run_cell(info_object_code)
        
    def _update_function_info_object (self):
        keys = dict(current_values_keys = list(self.current_values.keys()),
            previous_values_keys = list(self.previous_values.keys()),
            argument_variables=self.argument_variables,
            read_only_variables=self.read_only_variables,
            previous_variables=self.previous_variables,
            created_variables=self.created_variables)
        joblib.dump (keys, 'function_processor_keys.pk')
        
        #pdb.no_set_trace()
        keys_update_code=(
f'''
import joblib
keys = joblib.load ('function_processor_keys.pk')
{self.name}_info.current_values = {{k: {self.name}_info.current_values[k] for k in keys['current_values_keys']}}
{self.name}_info.previous_values = {{k: '__REMOVED__' for k in keys['previous_values_keys']}}
{self.name}_info.update ({{k: keys[k] for k in ['argument_variables', 'read_only_variables', 'previous_variables', 'created_variables']}})
''')
        get_ipython().run_cell(keys_update_code)

        
    def run_code_and_collect_locals (self, code=None, is_test_function=False, store_values=True):
        ##pdb.no_set_trace()
        if code is None: code=self.original_code
        
        joblib.dump (dict(self), 'function_processor.pk')
        if not is_test_function:
            #pdb.no_set_trace()
            self._store_values ("previous_values", code="", store_values=store_values)
            self._store_values ("current_values", code=code + "\n", store_values=store_values)
        else:
            get_ipython().run_cell ('from nbmodular.core.cell2func import get_non_callable_ipython\nget_non_callable_ipython ("previous_variables", locals())')
            get_ipython().run_cell ('from nbmodular.core.cell2func import get_non_callable_ipython\nget_non_callable_ipython ("created_variables", locals())')
            self.previous_values = {k: None for k in self.previous_variables}
            self.current_values = {k: None for k in self.created_variables}
        
        self._create_function_info_object ()
        self.match_variables_and_locals ()
        
    def match_variables_and_locals (self):
        #pdb.no_set_trace()
        # previous variables / values
        self.previous_variables = [k for k in self.previous_variables if k in self.previous_values]
        self.previous_variables += [k for k in self.argument_variables if k in self.previous_values and k not in self.previous_variables]
        self.previous_variables += [k for k in self.created_variables if k in self.previous_values and k in self.loaded_names+self.argument_variables and k not in self.previous_variables]
        self.previous_values = {k:self.previous_values[k] for k in self.previous_values if k in self.previous_variables}
        
        # created variables / current values
        self.current_values = {k:self.current_values[k] for k in self.current_values if k in self.created_variables or k in self.return_values}
        self.all_values = {**self.previous_values, **self.current_values}
        self.argument_variables = [k for k in self.argument_variables if k in self.all_values]
        self.read_only_variables = [k for k in self.read_only_variables if k in self.all_values]
        
        self.all_variables = self.created_variables.copy()
        self.all_variables += [k for k in self.previous_variables if k not in self.all_variables]
        self.all_variables += [k for k in self.argument_variables if k not in self.all_variables]
        
        self.current_values = {k:self.current_values[k] for k in self.current_values if k in self.all_variables}
        self.previous_values = {k:self.previous_values[k] for k in self.previous_values if k in self.all_variables}
        self.logger.debug (f'Stored the following local variables in the {self.name} current_values dictionary: {list(self.current_values.keys())}')
        
    def add_function_call (self, function):
        if 'added_functions' not in self:
            self.added_functions = []
        if 'function_calls' not in self:
            self.function_calls = '' 
        if function.name not in self.added_functions:
            self.added_functions.append (function.name)
            self.function_calls += f'{" "*self.tab_size}' + ','.join (function.return_values) + f' = {function.name}()\n'
            
    def add_to_signature (self, input=None, output=None, **kwargs):
        if input is not None:
            self.arguments += input
        if output is not None:
            self.return_values  += output
        if input is not None or output is not None:
            self.signature=None
            self.update_code ()
    
    def __str__ (self):
        name = None if not hasattr(self, 'name') else self.name
        current_values = self.current_values.keys() if hasattr(self, 'current_values') and self.current_values is not None else None
        return f'FunctionProcessor with name {name}, and fields: {self.keys()}\n    Arguments: {self.arguments}\n    Output: {self.return_values}\n    Locals: {current_values}'
    
    def __repr__ (self):
        return str(self)
    
    def store_variables (self, path_variables):
        store_variables_code = f'\nfrom nbmodular.core.cell2func import store_variables\nstore_variables ("{path_variables}", locals ())'
        get_ipython().run_cell(store_variables_code)
        
    def get_cell_function_code (self, unique=True, first=False, last=False):
        code = self.code
        if unique:
            function_code = function.code
        elif first:
            function_code = f'{self.code_parts.signature}\n' + self.code_parts.function_calls + self.code_parts.unpack_input_code + self.code_parts.function_code
        elif last:
            function_code = self.code_parts.function_code + self.code_parts.pack_output_code + self.code_parts.return_line
        else:
            function_code = self.code_parts.function_code
        return function_code

# %% ../../nbs/cell2func.ipynb 13
def update_cell_code (
    cell, 
    defined=False
):
    if defined:
        v = cell.split(':')
        signature = v[0] 
        signature += ':'
        cell=':'.join(v[1:])
    else:
        signature = None
    
    original_code = ''
    for line in cell.splitlines():
        if 'return' not in line and '#|' not in line:
            original_code += (line + '\n')
    cell = 'if True:\n' + original_code if defined else original_code
    return cell, signature

# %% ../../nbs/cell2func.ipynb 15
def add_function_to_list (function, function_list, idx=None):
    if idx is None:
        function_list.append (function)
    else:
        function_list[idx] = function
    return function_list

# %% ../../nbs/cell2func.ipynb 17
class CellProcessor():
    """
    Processes the cell's code according to the magic command.
    """
    def __init__(
        self, 
        tab_size=4, 
        log_level='INFO',
        restrict_inputs=False,
        **kwargs,
    ):
        self.logger = logging.getLogger('CellProcessor')
        self.logger.setLevel(log_level)
        
        self.restrict_inputs = restrict_inputs
        self.function_info = Bunch()
        self.current_function = Bunch()
        self.function_list = []
        
        self.test_function_info = Bunch()
        self.test_function_list = []
        
        self.test_data_function_info = Bunch()
        self.test_data_function_list = []
        
        self.all_variables = set()
        self.test_data_all_variables = set()
        self.test_all_variables = set()

        self.cell_nodes = []        
        self.cell_nodes_per_function = Bunch()
        
        self.imports = ''
        #self.test_imports = 'from sklearn.utils import Bunch\nfrom pathlib import Path\nimport joblib\nimport pandas as pd\nimport numpy as np\n'
        self.test_imports = ''
        
        self.tab_size=tab_size
        try:
            self.file_name = ipynbname.name().replace ('.ipynb', '.py')
            try:
                x = int (self.file_name[0])
                self.file_name = f'n{self.file_name}'
            except ValueError:
                pass
            nb_path = ipynbname.path ()
            found_notebook = True
        except FileNotFoundError:
            self.file_name = 'temporary.py'
            nb_path = Path ('.').absolute()
            found_notebook = False
        
        self.file_name_without_extension = self.file_name.split('.')[0]
        
        try:
            self.nbs_folder = self.get_nbs_path ()
            self.lib_folder = self.get_lib_path ()
        except:
            found_notebook = False
        
        if found_notebook:
            ##pdb.no_set_trace()
            try:
                index = nb_path.parts.index(self.nbs_folder.name)
            except:
                index = -1
            self.file_path = (self.nbs_folder.parent / self.lib_folder.name).joinpath (*nb_path.parts[index+1:])
            self.file_path = self.file_path.parent / self.file_path.name.replace ('.ipynb', '.py')
            self.file_path.parent.mkdir (parents=True, exist_ok=True)
            if index > -1:
                self.test_file_path = (self.nbs_folder.parent / 'tests').joinpath (*nb_path.parts[index+1:-1])/ f'test_{self.file_path.name}'
            else:
                self.test_file_path = self.file_path.parent / f'test_{self.file_path.name}'
            self.test_file_path.parent.mkdir (parents=True, exist_ok=True)
        else:
            file_name = self.file_name.replace ('.ipynb', '.py')
            self.file_path = nb_path / file_name
            self.test_file_path = nb_path /  f'test_{file_name}'
            
        self.call_history = []
        self.load_tests = True
        self.save_tests = True
        self.run_tests = True
        
        self.parser = argparse.ArgumentParser(description='Process some integers.')
        self.parser.add_argument('-i', '--input', type=str, nargs='+', help='Strict input. No other input considered, regardless of any dependencies.')
        self.parser.add_argument('--include-input', type=str, default=[], nargs='+', help='Input to be included in addition of identified dependencies on previous functions.')
        self.parser.add_argument('--exclude-input', type=str, default=[], nargs='+', help='Input to be excluded.')
        self.parser.add_argument('-o', '--output', type=str, nargs='+', help='Strict output. No other output considered, regardless of any dependencies.')
        self.parser.add_argument('--include-output', type=str, default=[], nargs='+', help='Output to be included in addition of identified dependencies from posterior functions')
        self.parser.add_argument('--exclude-output', type=str, default=[], nargs='+', help='Output to be excluded.')
        self.parser.add_argument('-m', '--merge',  action='store_true', help='merge with previous function')
        self.parser.add_argument('-s', '--show',  action='store_true', help='show function code')
        self.parser.add_argument('-l', '--load',  action='store_true', help='load variables')
        self.parser.add_argument('--save',  action='store_true', help='save variables')
        self.parser.add_argument('-n', '--not-run',  action='store_true', help='do not execute the contents of the cell')
        self.parser.add_argument('--not-store',  action='store_true', help='do not store local values from cell')
        self.parser.add_argument('--not-store-locals-in-disk',  action='store_true', help='do not store local values from cell in disk')
        self.parser.add_argument('--override',  action='store_true', help='load / save / no-run values override any global flags')
        self.parser.add_argument('--returns-dict',  action='store_true', help='function results are gathered in dictionary' )
        self.parser.add_argument('--returns-bunch',  action='store_true', help='function results are gathered in Bunch' )
        self.parser.add_argument('--unpack-bunch',  type=str, default=None, help='uses bunch variable as I/O, and unpacks its contents at beginning of function' )
        self.parser.add_argument('--pack-all',  action='store_true', help='pack all variables in bunch, including those not recognized as variables created or modified in function')
        self.parser.add_argument('--only-posterior',  action='store_true', help='return only created / modified variables that are used in posterior functions')
        self.parser.add_argument('--return-all',  action='store_true', help='return all created / modified variables even if they are not used in posterior functions')
        self.parser.add_argument('-t', '--test',  action='store_true', help='test function / imports')
        self.parser.add_argument('-d', '--data',  action='store_true', help='data function')
        self.parser.add_argument('-p', '--permanent',  action='store_true', help='do not change the contents of the function')
        self.parser.add_argument('--name', type=str, help='name of function to debug')
        self.parser.add_argument('--idx', type=int, help='position of function to debug')
        self.parser.add_argument('--history',  action='store_true', help='resets everything, including history')

    def set_load_tests (self, value):
        if (value != self.load_tests):
            self.logger.info (f'changing global load flag to {value}')
        self.load_tests = value
    def set_save_tests (self, value):
        if (value != self.save_tests):
            self.logger.info (f'changing global save flag to {value}')
        self.save_tests = value
    def set_run_tests (self, value):
        if (value != self.run_tests):
            self.logger.info (f'changing global run flag to {value}')
        self.run_tests = value
        
    def debug_function (self, call_history=None, idx=None, name=None, test=False, data=False, **kwargs):
        if call_history is not None:
            self.call_history = call_history

        if name is not None:
            function_info = (self.function_info if not test else
                             self.test_function_info if test and not data else
                             self.test_data_function_info)
            if name not in function_info:
                idx = 0
                for func, code in self.call_history:
                    if func==name:
                        break
                    idx += 1
            else:
                idx = function_info[name].idx
        elif idx is None:
            idx = -1
            
        self.reset()
            
        for call in self.call_history[:idx]:
            self.process_function_call (*call, add_call=False)

        import ipdb
        ipdb.runcall (self.process_function_call, *self.call_history[idx], add_call=False)
        
    def reset_function_list (self, function_list, function_info):
        if len(function_list)==0:
            return [], Bunch()
        values_to_remove = [x for function in function_list for x in function.current_values.keys()]
        remove_variables_code = '\n'.join([f'''
            try:
                exec("del {x}")
            except:
                print (f'could not remove {x}')
                ''' for x in values_to_remove])
        get_ipython().run_cell(remove_variables_code)
        return [], Bunch()
        
    def reset (self, remove_history=False, **kwargs):
        self.function_list, self.function_info = self.reset_function_list (self.function_list, self.function_info)
        self.test_function_list, self.test_function_info = self.reset_function_list (self.test_function_list, self.test_function_info)
        self.test_data_function_list, self.test_data_function_info = self.reset_function_list (self.test_data_function_list, self.test_data_function_info)
        
        self.all_variables = set()
        self.test_data_all_variables = set()
        self.test_all_variables = set()

        self.imports = ''
        self.test_imports = ''
        if remove_history:
            self.call_history = []
    
    def process_function_call (self, line, cell, add_call=True):
        call = (line, cell)
        if add_call:
            self.add_call (call)
        function_name, kwargs = self.parse_signature (line)
        self.function (function_name, cell, call=call, original_kwargs=kwargs.copy(), **kwargs)

    def add_call (self, call):
        self.call_history.append (call)
                   
    def create_function (
        self, 
        cell, 
        func, 
        call,
        original_kwargs=None,
        unknown_input=None,
        unknown_output=None,
        test=False,
        data=False,
        permanent=False,
        not_run=False,
        returns_dict=False,
        returns_bunch=False,
        unpack_bunch=None,
        return_all=False,
        only_posterior=False,
        include_input=[],
        exclude_input=[],
        include_output=[],
        exclude_output=[],
        not_store_locals_in_disk=False,
        cell_idx=None,
        **kwargs,
    ):
        #pdb.no_set_trace()
        original_name=func
        root = ast.parse (cell)
        name=[x.name for x in ast.walk(root) if isinstance (x, ast.FunctionDef)]
            
        if len(name)>0:
            if len(name) > 0:
                name=name[0]
                func=name
            arguments = [[x.arg for x in node.args.args] for node in ast.walk(root) if isinstance (node, ast.FunctionDef)]
            if len(arguments)>0:
                arguments = arguments[0]
            return_values = [([x.id for x in node.value.elts if isintance (node, ast.Name)] if hasattr(node.value, 'elts') else [node.value.id]) if isinstance (node.value, ast.Name) else '' for node in ast.walk(root) if isinstance (node, ast.Return)] 
            if len(return_values)>0:
                return_values = return_values[0]
            
            unknown_input = False
            unknown_output = False
            input = arguments
            output = return_values
            defined = True
            joblib.dump (cell, 'cell.pk')
            argument_initialization_code=(
f'''
import ast
import joblib
import os

def get_args_and_defaults (list_args, list_defaults):
    default_values =  [y.value for y in list_defaults]
    arguments = [arg.arg for arg in list_args[len(list_defaults):]]
    return arguments, default_values
    
cell = joblib.load ('cell.pk')
os.remove ('cell.pk')
root = ast.parse (cell)

args_with_defaults1, default_values1 =  get_args_and_defaults (root.body[0].args.posonlyargs + root.body[0].args.args, root.body[0].args.defaults)
args_with_defaults2, default_values2 =  get_args_and_defaults (root.body[0].args.kwonlyargs, root.body[0].args.kw_defaults)

for arg, val in zip (args_with_defaults1+args_with_defaults2, default_values1+default_values2):
    exec (arg + "=" + str(val))
''')
            get_ipython().run_cell(argument_initialization_code)
        else:
            defined = False
            arguments=[]
            return_values=[]
        
        if defined and not permanent:
            cell, signature = update_cell_code (cell, True)
        else:
            cell, signature = update_cell_code (cell, False)
            
        if returns_bunch:
            self.imports += 'from sklearn.utils import Bunch\n'
        
        this_function = FunctionProcessor (
            original_code=cell, 
            original_cell=cell,
            original_name=original_name,
            name=func, 
            call=call,
            tab_size=self.tab_size,
            arguments=arguments,
            return_values=return_values,
            unknown_input=unknown_input,
            unknown_output=unknown_output,
            test=test,
            data=data,
            defined=defined,
            permanent=permanent,
            signature=signature,
            not_run=not_run,
            previous_values={},
            current_values={},
            returns_dict=returns_dict,
            returns_bunch=returns_bunch,
            unpack_bunch=unpack_bunch,
            include_input=include_input,
            exclude_input=exclude_input,
            include_output=include_output,
            exclude_output=exclude_output,
            store_locals_in_disk=not not_store_locals_in_disk,
            original_kwargs=original_kwargs,
            cell_idx=cell_idx,
            logger=self.logger,
        )
        if defined and permanent:
            this_function.code = cell
        this_function.parse_variables ()
        
        if False and defined:
            loaded_names_not_in_arguments = set(this_function.loaded_names).difference (arguments)
            if len(loaded_names_not_in_arguments) > 0:
                self.logger.debug (f'The following loaded names were not found in the arguments list: {loaded_names_not_in_arguments}')
        
        return this_function
    
    def create_function_and_run_code(
        self,
        func, 
        cell,
        call=None,
        input=None,
        unknown_input=True,
        output=None,
        unknown_output=True,
        not_store=False,
        make_function=True,
        update_previous_functions=True,
        show=False,
        load=False,
        save=False,
        not_run=False,
        override=False,
        test=False,
        data=False,
        permanent=False,
        restrict_inputs=None,
        idx=None,
        **kwargs,
    ) -> FunctionProcessor:
        
        if restrict_inputs is None:
            restrict_inputs = self.restrict_inputs
        store_values = not not_store
        #pdb.no_set_trace()
        if test:
            func = 'test_' + func
            
        if test and not data and not override:
            load = self.load_tests
            save = self.save_tests
            not_run = not self.run_tests
        
        self.current_function = self.create_function (
            cell, 
            func, 
            call, 
            unknown_input=unknown_input,
            unknown_output=unknown_output,
            test=test,
            data=data,
            permanent=permanent,
            not_run=not_run,
            **kwargs
        )
        
        # register
        if idx is None:
            idx = self.current_function.idx = len(self.function_list)
        
        # get variables specific about this function
        path_variables = Path (self.file_name_without_extension) / f'{func}.pk'
        if load and path_variables.exists():
            self.current_function.store_variables (path_variables)
            not_run=True

        if not not_run:
            is_test_function = self.current_function.test and not self.current_function.data
            self.current_function.run_code_and_collect_locals(is_test_function=is_test_function, store_values=store_values)
                    
        if make_function:
            # TODO: what if the current function is self.current_function.data is True?
            arguments=(self.current_function.previous_variables if unknown_input and not self.current_function.test and not self.current_function.defined else 
                           [] if self.current_function.test else
                           self.current_function.arguments if self.current_function.defined else
                           input)
            
            if not self.current_function.test and not self.current_function.data:
                variables_created_by_previous_functions = [x for f in self.function_list[:idx] for x in f.created_variables]
                if self.restrict_inputs:
                    # arguments can only be variables created by previous functions ("created_variables") 
                    # that are loaded in the current one. If only stored, they cannot be in arguments.
                    arguments=[
                        x for x in arguments if (x in self.current_function.loaded_names 
                                                 and x in variables_created_by_previous_functions)
                    ]
                    self.current_function.previous_variables = [
                        x for x in self.current_function.previous_variables if (x in self.current_function.loaded_names 
                                                 and x in variables_created_by_previous_functions)
                    ]
                else:
                    # arguments can only be variables that are either:
                    # - created by previous functions ("created_variables") 
                    # - not created by previous functions, but not created by either the current function or subsequent functions either.
                    variables_created_by_this_or_posterior_functions = [x for f in self.function_list[idx:] for x in f.created_variables]
                    arguments=[
                        x for x in arguments if (x in self.current_function.loaded_names 
                                                 and (x in variables_created_by_previous_functions or 
                                                      x not in variables_created_by_this_or_posterior_functions))
                    ]
                    self.current_function.previous_variables = [
                        x for x in self.current_function.previous_variables if (x in self.current_function.loaded_names 
                                                                                and (x in variables_created_by_previous_functions or 
                                                                                     x not in variables_created_by_this_or_posterior_functions))
                    ]
            
            # return values
            return_values=([] if unknown_output and not self.current_function.defined else 
                               self.current_function.return_values if self.current_function.defined else 
                               output)
            # update code based on those
            self.current_function.update_code ( 
                arguments=arguments, 
                return_values=return_values,
                display=show
            )
            
        # add variables from current function to posterior_variables of all the previous functions
        ##pdb.no_set_trace()
        # test data functions have output dependencies on test functions
        # test functions have no output dependencies
        function_list = (self.function_list if not self.current_function.test and not self.current_function.data 
                         else self.test_data_function_list if self.current_function.test and not self.current_function.data
                         else [])
        if not self.current_function.test and not self.current_function.data:
            function_list = function_list[:idx]
        # if test function, its input comes from test data functions: 
        # - 1 add output dependencies to test data functions
        # - 2 add input dependencies from each test data function
        #if self.current_function.test:
        #    #pdb.no_set_trace()
        for function in function_list:
            function.posterior_variables += [v for v in self.current_function.previous_variables if v not in function.posterior_variables]
            if update_previous_functions and unknown_output and not function.defined:
                function.update_code (
                    return_values=[x for x in function.created_variables + function.argument_variables if x in function.posterior_variables], 
                    display=False
                )
            if self.current_function.test and function.test and function.data:
                self.current_function.add_function_call (function)
        
        if self.current_function.test and not self.current_function.data and not not_run: 
            self.current_function.run_code_and_collect_locals(is_test_function=False, store_values=store_values)
        
        if self.current_function.test and not self.current_function.data:
            self.current_function.update_code()
            
        if self.current_function.defined:
            potential_arguments = self.current_function.loaded_names
            #potential_arguments = self.current_function.previous_variables
            potential_arguments = list (set (potential_arguments).difference (self.current_function.arguments))
            self.posterior_not_in_results = list (set (self.current_function.posterior_variables).difference (self.current_function.return_values))
            
            non_callable = get_non_callable (potential_arguments)
            if len (non_callable) > 0:
                self.logger.info (f'Detected the following previous variables that are not in the argument list: {non_callable}')
            if len (self.posterior_not_in_results) > 0:
                self.logger.info (f'Detected the following posterior variables that are not in the return list: {self.posterior_not_in_results}')
        
        if self.current_function.test and self.current_function.data:
            common = set(self.current_function.all_variables).intersection (self.test_data_all_variables)
            if len(common)>0:
                raise ValueError (f'detected common variables with other test data functions {common}:')
        
        if not self.current_function.test and not self.current_function.data:
            self.all_variables |= set(self.current_function.all_variables)
        elif self.current_function.test and self.current_function.data:
            self.test_data_all_variables |= set(self.current_function.all_variables)
        elif self.current_function.test and not self.current_function.data:
            self.test_all_variables |= set(self.current_function.all_variables)
        
        self.current_function._update_function_info_object ()
        
        if save and not not_run:
            path_variables.parent.mkdir (parents=True, exist_ok=True)
            joblib.dump (self.current_function.current_values, path_variables)
        
        return self.current_function
    
    def function (
        self,
        func,
        cell,
        merge=False,
        show=False,
        register_pipeline=True,
        pipeline_name=None,
        write=True,
        test=False,
        data=False,
        **kwargs
    ) -> None:

        function_list = (self.test_function_list if test and not data else 
                         self.test_data_function_list if test and data else 
                         self.function_list)
        func_name = f'test_{func}' if test else func
        existing = False
        for idx, f in enumerate(function_list):
            if f.name == func_name:
                existing=True
                #function_list.remove (f)
                break
        if not existing:
            idx = None
            
        #pdb.no_set_trace()
        this_function = self.create_function_and_run_code(func, cell, show=show, test=test, data=data, idx=idx, **kwargs)
        if existing and (func in self.function_info) and merge:
            this_function = self.merge_functions (self.function_info[func], this_function, idx=idx, show=show)
            function_list[idx] = this_function
        
        function_name = this_function.name
        if this_function.test:
            if this_function.data:
                self.test_data_function_info[function_name] = this_function
                self.test_data_function_list[idx] = this_function
                self.test_data_function_list = add_function_to_list (this_function, self.test_data_function_list, idx=idx)
            else:
                self.test_function_info[function_name] = this_function
                self.test_function_list = add_function_to_list (this_function, self.test_function_list, idx=idx)
        else:
            self.function_info[function_name] = this_function
            self.function_list = add_function_to_list (this_function, self.function_list, idx=idx)
       
        if this_function.name in self.cell_nodes_per_function:
            self.cell_nodes_per_function[this_function.name].append(this_function.copy())
        else:
            self.cell_nodes_per_function[this_function.name] = [this_function.copy()]
            
        self.cell_nodes.append(this_function)
        self.export_cell_nodes ()
            
        if register_pipeline:
            self.register_pipeline (pipeline_name=pipeline_name)
        else:
            self.pipeline = None
        if write:
            self.write ()
            self.write (test=True)
            
    def export_cell_nodes (self):
        for cell_function in self.cell_nodes:
            #code = cell_function.get_cell_function_code (unique=len(nodes_per_function[cell_function.name])==1,
            #                                             first=cell_function is nodes_per_function[cell_function.name][0],
            #                                             last=cell_function is nodes_per_function[cell_function.name][-1])
            # make_cell (cell=code)
            pass

            
    def merge_functions (self, this_function, new_function, idx=None, show=False):
        
        this_function.original_code += new_function.original_code
        original_kwargs = {**this_function.original_kwargs, **new_function.original_kwargs}
        
        this_function = self.create_function_and_run_code(
            this_function.original_name,
            this_function.original_code,
            this_function.call,
            original_kwargs=original_kwargs,
            idx=idx,
            **original_kwargs,
        )
        
        return this_function
    
    def parse_args (self, line):
        argv = shlex.split(line, posix=(os.name == 'posix'))
        pars = self.parser.parse_args(argv)
        kwargs = vars(pars)
        return kwargs
                
    def parse_signature (self, line):
        argv = shlex.split(line, posix=(os.name == 'posix'))
        
        function_name=argv[0] if len(argv)>0 else ''
        signature = dict(
            input=None,
            unknown_input=True,
            output=None,
            unknown_output=True
        )
        found_io = False
        for idx, arg in enumerate(argv):
            if arg and arg.startswith('-') and arg != '-' and arg != '->':
                found_io = True
                break
        if found_io and idx==0:
            function_name = ''
            ##pdb.no_set_trace()
        if found_io:
            pars = self.parser.parse_args(argv[idx:])
            unknown_input = 'input' not in pars
            if not unknown_input:
                signature.update (input=() if pars.input==['None'] else pars.input, unknown_input=pars.input is None)
            unknown_output = 'output' not in pars
            if not unknown_output:
                signature.update (output=() if pars.output==['None'] else pars.output, unknown_output=pars.output is None)
            kwargs = vars(pars)
        else:
            kwargs = {}
        kwargs.update (signature)
            
        return function_name, kwargs
    
    def write_imports (
        self,
        cell,
        test=False,
        **kwargs
    ):
        get_ipython().run_cell (cell)
        if not test:
            self.imports += cell
        else:
            self.test_imports += cell
        self.write (test=test)
    
    def write (self, test=False):
        ##pdb.no_set_trace()
        function_list = self.function_list if not test else self.test_function_list
        file_path = self.file_path if not test else self.test_file_path
        imports = self.imports if not test else self.test_imports
        with open (str(file_path), 'w') as file:
            ##pdb.no_set_trace()
            file.write (imports)
            for function in function_list:
                function.write (file)
            if not test and self.pipeline is not None:
                self.pipeline.write (file)
                
    def print (self, function_name, test=False, data=False, **kwargs):
        if function_name == 'all':
            function_list = self.test_data_function_list if test and data else self.test_function_list if test else self.function_list
            for function in function_list:
                function.print ()
        else:
            if test and data:
                self.test_data_function_info[function_name].print ()
            elif test:
                self.test_function_info[function_name].print ()
            else:
                self.function_info[function_name].print ()
            
    def get_lib_path (self):
        return nbdev.config.get_config()['lib_path']
                   
    def get_nbs_path (self):
        return nbdev.config.get_config()['nbs_path']
    
    def pipeline_code (self, pipeline_name=None):
        pipeline_name = f'{self.file_name_without_extension}_pipeline' if pipeline_name is None else pipeline_name
        
        code = (
f'''# -----------------------------------------------------
# pipeline
# -----------------------------------------------------
def {pipeline_name} (test=False, load=True, save=True, result_file_name="{pipeline_name}"):
    """Pipeline calling each one of the functions defined in this module."""
    
    # load result
    result_file_name += '.pk'
    path_variables = Path ("{self.file_name_without_extension}") / result_file_name
    if load and path_variables.exists():
        result = joblib.load (path_variables)
        return result

''')
        return_values = set()
        for func in self.function_list:
            argument_list_str = ", ".join(func.arguments) if not func.data else "test=test"
            return_list_str = f'{", ".join(func.return_values)} = ' if len(func.return_values)>0 else ''
            return_values |= set(func.return_values)
            code += f'{" " * self.tab_size}' + f'{return_list_str}{func.name} ({argument_list_str})\n'
        
        return_values = list (return_values)
        result_str = "Bunch (" + "".join([f"{k}={k}," for k in return_values[:-1]])
        if len(return_values)>0:
            k = return_values[-1] 
            result_str = result_str + f"{k}={k}" + ")"
        else:
            result_str = result_str + ")"
        
        code += (
f'''
    # save result
    result = {result_str}
    if save:    
        path_variables.parent.mkdir (parents=True, exist_ok=True)
        joblib.dump (result, path_variables)
    return result
''')
            
        return code, pipeline_name
    
    def test_pipeline_code (self, pipeline_name=None):
        pipeline_name = f'{self.file_name_without_extension}_pipeline' if pipeline_name is None else pipeline_name
        result_file_name_with_braces = 'f"test_{result_file_name}"'
        code = (
f'''
from sklearn.utils import Bunch
from pathlib import Path
import joblib
import pandas as pd
import numpy as np

def test_{pipeline_name} (test=True, prev_result=None, result_file_name="{pipeline_name}"):
    result = {pipeline_name} (test=test, load=True, save=True, result_file_name=result_file_name)
    if prev_result is None:
        prev_result = {pipeline_name} (test=test, load=True, save=True, result_file_name={result_file_name_with_braces})
    for k in prev_result:
        assert k in result
        if type(prev_result[k]) is pd.DataFrame:    
            pd.testing.assert_frame_equal (result[k], prev_result[k])
        elif type(prev_result[k]) is np.array:
            np.testing.assert_array_equal (result[k], prev_result[k])
        else:
            assert result[k]==prev_result[k]
''')
        return code, f'test_{pipeline_name}'
    
    def register_pipeline (self, pipeline_name=None):
        code, name = self.pipeline_code (pipeline_name=pipeline_name)
        self.pipeline = FunctionProcessor (code=code,
                                           arguments=[],
                                           return_values=[],
                                           name=name)
        get_ipython().run_cell(self.imports)
        get_ipython().run_cell(code)
        
        code, name = self.test_pipeline_code (pipeline_name=pipeline_name)
        self.test_pipeline = FunctionProcessor (code=code,
                                           arguments=[],
                                           return_values=[],
                                           name=name)
        get_ipython().run_cell(self.test_imports)
        get_ipython().run_cell(code)
    
    def print_pipeline (self, test=False, **kwargs):
        if test:
            code, name = self.test_pipeline_code()  
        else:
            code, name = self.pipeline_code()  
        print (code)

# %% ../../nbs/cell2func.ipynb 19
@magics_class
class CellProcessorMagic (Magics):
    """
    Base magic class for converting cells to modular functions.
    """
    def __init__(self, shell, **kwargs):
        super().__init__(shell)
        self.processor = CellProcessor (magic=self, **kwargs)
        
    @cell_magic
    def cell2file (self, folder, cell):
        self.processor.cell2file (folder, cell)
    
    @cell_magic
    def cell_text (self, line, cell):
        "Converts cell to function"
        return cell
        
    @cell_magic
    def function (self, line, cell):
        "Converts cell to function"
        self.processor.process_function_call (line, cell)
        
    @cell_magic
    def debug_cell (self, line, cell):
        """
        Debugs cell as if it had a `function` magic command.
        
        %%debug_cell <my_function>
        (function_code)
        """
        import ipdb
        ipdb.runcall (self.processor.process_function_call, line, cell)
        
    @cell_magic
    def imports (self, line, cell):
        "Converts cell to function"
        kwargs = self.processor.parse_args (line)
        self.processor.write_imports (cell, **kwargs)
    
    @line_magic
    def write (self, line):
        return self.processor.write ()
    
    @line_magic
    def print (self, line):
        ##pdb.no_set_trace()
        function_name, kwargs = self.processor.parse_signature (line)
        return self.processor.print (function_name, **kwargs)
    
    @line_magic
    def function_info (self, line):
        function_name, kwargs = self.processor.parse_signature (line)
        ##pdb.no_set_trace()
        if kwargs.get('test', False):
            return self.processor.test_function_info [function_name]
        else:
            return self.processor.function_info [function_name]
        
    @line_magic
    def add_to_signature (self, line):
        function_name, kwargs = self.processor.parse_signature (line)
        self.processor.function_info[function_name].add_to_signature (**kwargs)
    
    @line_magic
    def cell_processor (self, line):
        return self.processor
        
    @line_magic
    def pipeline_code (self, line):
        return self.processor.pipeline_code ()
    
    @line_magic
    def print_pipeline (self, line):
        kwargs = self.processor.parse_args (line)
        return self.processor.print_pipeline (**kwargs)
          
    @line_magic
    def match (self, line):
        p0 = '[a-zA-Z]\S*\s*\\([^-()]*\\)\s*->\s*\\([^-()]*\\)'
        p = '\\([^-()]*\\)'
        m = re.search (p0, line)
        if m is not None:
            inp, out = re.findall (p, line)
            #print (inp)
            #print (out)
            
    @line_magic
    def debug_function (self, line):
        kwargs = self.processor.parse_args (line)
        self.processor.debug_function (**kwargs)
        
    @line_magic
    def reset (self, line):
        kwargs = self.processor.parse_args (line)
        self.processor.reset (**kwargs)
        
    @line_magic
    def load_tests (self, line):
        self.processor.set_load_tests (True)
        
    @line_magic
    def load_tests (self, line):
        self.processor.set_load_tests (True)

    @line_magic
    def not_load_tests (self, line):
        self.processor.set_load_tests (False)

    @line_magic
    def save_tests (self, line):
        self.processor.set_save_tests (True)

    @line_magic
    def not_save_tests (self, line):
        self.processor.set_save_tests (False)

    @line_magic
    def run_tests (self, line):
        self.processor.set_run_tests (True)

    @line_magic
    def not_run_tests (self, line):
        self.processor.set_run_tests (False)

# %% ../../nbs/cell2func.ipynb 21
def load_ipython_extension(ipython):
    """
    This module can be loaded via `%load_ext core.cell2func` or be configured to be autoloaded by IPython at startup time.
    """
    magics = CellProcessorMagic(ipython)
    ipython.register_magics(magics)

# %% ../../nbs/cell2func.ipynb 23
def keep_variables (field, variable_values, self=None):
    """
    Store `variables` in disk
    """
    import joblib
    import os
    variable_values = {k: variable_values[k] for k in variable_values if acceptable_variable(variable_values, k)}
    joblib.dump (variable_values, 'variable_values.pk')

# %% ../../nbs/cell2func.ipynb 24
def keep_variables_in_memory (field, variable_values, self=None):
    """
    Store `variables` in dictionary entry `self[field]`
    """
    frame_number = 0
    ##pdb.no_set_trace()
    while not isinstance (self, FunctionProcessor):
        try:
            fr = sys._getframe(frame_number)
        except:
            break
        args = argnames(fr, True)
        if len(args)>0:
            self = fr.f_locals[args[0]]
        frame_number += 1
    if isinstance (self, FunctionProcessor):
        variable_values = {k: variable_values[k] for k in variable_values if acceptable_variable(variable_values, k)}
        variable_values['created_current_values'] = True
        self[field]=variable_values.copy()
        #del variable_values['created_current_values']

# %% ../../nbs/cell2func.ipynb 25
def acceptable_variable (variable_values, k):
    return not k.startswith ('_') and not callable(variable_values[k]) and type(variable_values[k]).__name__ not in ['module', 'FunctionProcessor', 'CellProcessor'] and k not in ['variable_values', 'In', 'Out']

# %% ../../nbs/cell2func.ipynb 27
import pdb
def store_variables (path_variables, locals_, self=None):
    """
    Store `variables` in dictionary entry `self.variables_field[function]`
    """
    ##pdb.no_set_trace()
    import joblib
    current_values = joblib.load (path_variables)
    #v = locals()
    #v.update (current_values)
    locals_.update (current_values)
    
    
    frame_number = 0
    ##pdb.no_set_trace()
    while not isinstance (self, FunctionProcessor):
        try:
            fr = sys._getframe(frame_number)
        except:
            break
        args = argnames(fr, True)
        if len(args)>0:
            self = fr.f_locals[args[0]]
        frame_number += 1
        
    if isinstance (self, FunctionProcessor):
        self['current_values']=current_values
    
